local ADDON_NAME = ...;
local DEFAULT_SAVED_VARS = { isShown = false, commandHistory = {}, messageHistory = {}, height = 300, };
local SAVED_VARS_VERSION = 2;

local MAX_NUM_COMMAND_HISTORY = 100;
local MAX_NUM_MESSAGE_HISTORY = 1000;

DeveloperConsoleMixin = {};

function DeveloperConsoleMixin:OnLoad()
	self:RegisterEvent("ADDON_LOADED");
	self:RegisterEvent("ADDONS_UNLOADING");
	self:RegisterEvent("TOGGLE_CONSOLE");
	self:RegisterEvent("CONSOLE_MESSAGE");
	self:RegisterEvent("CONSOLE_CLEAR");
	self:RegisterEvent("CONSOLE_COLORS_CHANGED");
	self:RegisterEvent("DEBUG_MENU_TOGGLED");

	self.MessageFrame:SetMaxLines(MAX_NUM_MESSAGE_HISTORY);

	self.savedVars = Blizzard_Console_SavedVars;

	self.commandCircularBuffer = CreateCircularBuffer(MAX_NUM_COMMAND_HISTORY);
	self:ResetCommandHistoryIndex();

	self.MessageFrame:SetOnScrollChangedCallback(function(messageFrame, offset)
		messageFrame.ScrollBar:SetValue(messageFrame:GetNumMessages() - offset);
	end);

	self.filterText = "";
end

function DeveloperConsoleMixin:RestoreMessageHistory()
	local messageHistory = self.savedVars.messageHistory;
	if #messageHistory > 0 then
		self.savedVars.messageHistory = {};

		local numElements = math.min(MAX_NUM_MESSAGE_HISTORY, #messageHistory);

		for i = (#messageHistory - numElements) + 1, numElements do
			local message, colorType = unpack(messageHistory[i]);
			local color = C_Console.GetColorFromType(colorType);
			local r, g, b = color:GetRGB();
			self.MessageFrame:AddMessage(message, r, g, b, colorType);
			table.insert(self.savedVars.messageHistory, messageHistory[i]);
		end

		local sides = ("-"):rep(50);
		local previousSessionLine = ("%s Previous Session %s"):format(sides, sides);

		self:AddMessage(previousSessionLine, Enum.ConsoleColorType.DefaultColor);
	end
end

function DeveloperConsoleMixin:RestoreCommandHistory()
	local commandHistory = self.savedVars.commandHistory;
	if #commandHistory > 0 then
		self.savedVars.commandHistory = {};

		self.commandCircularBuffer:Clear();
		self:ResetCommandHistoryIndex();

		local numElements = math.min(MAX_NUM_COMMAND_HISTORY, #commandHistory);

		for i = (#commandHistory - numElements) + 1, numElements do
			self.commandCircularBuffer:PushFront(commandHistory[i]);
			table.insert(self.savedVars.commandHistory, commandHistory[i]);
		end
	end
end

function DeveloperConsoleMixin:OnEvent(event, ...)
	if event == "ADDON_LOADED" then
		if ADDON_NAME == ... then
			if not Blizzard_Console_SavedVars or not Blizzard_Console_SavedVars.version or Blizzard_Console_SavedVars.version < SAVED_VARS_VERSION then
				Blizzard_Console_SavedVars = CopyTable(DEFAULT_SAVED_VARS);
				Blizzard_Console_SavedVars.version = SAVED_VARS_VERSION;
			end

			self.savedVars = Blizzard_Console_SavedVars;

			if self.savedVars.isShown and not GetCVarBool("useNewConsole") then
				self.savedVars.isShown = false;
			end

			self:RestoreMessageHistory();
			self:RestoreCommandHistory();
			self:SetShown(self.savedVars.isShown);
			self:SetHeight(self.savedVars.height);
			self:UpdateAnchors();
		end
	elseif event == "ADDONS_UNLOADING" then
		local closingClient = ...;
		if closingClient then
			self.savedVars.isShown = false;
		end
	elseif event == "TOGGLE_CONSOLE" then
		local shownRequested = ...;
		self:Toggle(shownRequested);
	elseif event == "CONSOLE_MESSAGE" then
		local message, colorType = ...;

		self:AddMessage(message, colorType);
	elseif event == "CONSOLE_CLEAR" then
		self:Clear();
	elseif event == "CONSOLE_COLORS_CHANGED" then
		self:RefreshMessageFrame();
	elseif event == "DEBUG_MENU_TOGGLED" then
		self:UpdateAnchors();
	end
end

function DeveloperConsoleMixin:AddMessage(message, colorType)
	message = message:gsub("\n", "");

	if not colorType then
		colorType = Enum.ConsoleColorType.DefaultColor;
	end
	local color = C_Console.GetColorFromType(colorType);

	local r, g, b = color:GetRGB();
	self.MessageFrame:AddMessage(message, r, g, b, colorType);
	self:UpdateScrollbar();

	table.insert(self.savedVars.messageHistory, { message, colorType });
end

function DeveloperConsoleMixin:Clear()
	self.savedVars.messageHistory = {};
	self.MessageFrame:Clear();
end

function DeveloperConsoleMixin:RefreshMessageFrame()
	self.MessageFrame:Clear();

	local messageHistory = self.savedVars.messageHistory;
	for i, messageInfo in ipairs(messageHistory) do
		local message, colorType = unpack(messageInfo);
		local color = C_Console.GetColorFromType(colorType);
		local r, g, b = color:GetRGB();
		self.MessageFrame:AddMessage(message, r, g, b, colorType);
	end
end

function DeveloperConsoleMixin:CalculateAnchorOffset()
	if DebugMenu and DebugMenu.IsVisible() then
		return DebugMenu.GetMenuHeight();
	end
	return 0;
end

function DeveloperConsoleMixin:UpdateAnchors()
	local offset = self:CalculateAnchorOffset();
	self:SetPoint("TOPLEFT", 0, -offset);
	self:SetPoint("TOPRIGHT", 0, -offset);

	self:ValidateHeight();
	self.AutoComplete:OnAvailableHeightChanged();
end

function DeveloperConsoleMixin:OnMouseWheel(delta)
	if IsControlKeyDown() then
		delta = delta * self.MessageFrame:GetPagingScrollAmount(); 
	elseif IsShiftKeyDown() then
		delta = delta * 10;
	end 

	self.MessageFrame:ScrollByAmount(delta);
end

function DeveloperConsoleMixin:Toggle(shownRequested)
	if shownRequested == nil then -- toggle
		shownRequested = not self.savedVars.isShown;
	end

	if shownRequested ~= self.savedVars.isShown then
		self.savedVars.isShown = shownRequested;

		if shownRequested then
			self:Show();
		else
			self.EditBox:ClearFocus();
		end
	
		self.Anim.Translation:SetOffset(0, self.savedVars.height);
		self.Anim:Play(shownRequested);
	end
end

function DeveloperConsoleMixin:OnEscapePressed()
	if self.AutoComplete:IsShown() then
		self.AutoComplete:ForceHide();
		return;
	end
	self:Toggle(false);
end

function DeveloperConsoleMixin:UpdateScrollbar()
	local numMessages = self.MessageFrame:GetNumMessages();
	self.MessageFrame.ScrollBar:SetMinMaxValues(1, numMessages);
	self.MessageFrame.ScrollBar:SetValue(numMessages - self.MessageFrame:GetScrollOffset());
end

function DeveloperConsoleMixin:ValidateHeight(newHeight)
	local screenHeight = 768;
	local newHeight = Clamp(newHeight or self:GetHeight(), screenHeight * .1 + self.EditBox:GetHeight(), screenHeight * .85 - self:CalculateAnchorOffset());
	self:SetHeight(newHeight);
	if self.savedVars.height ~= newHeight then
		self.savedVars.height = newHeight;
		self.AutoComplete:OnAvailableHeightChanged();
	end
end

function DeveloperConsoleMixin:StartDragResizing()
	local startX, startY = GetCursorPosition();
	local startHeight = self:GetHeight();

	self.resizingTicker = C_Timer.NewTicker(0, function()
		local x, y = GetCursorPosition();
		self:ValidateHeight(startHeight + (startY - y));

		if not IsMouseButtonDown() then -- possible to get into this state when another window steals focus, should look at detecting that at the input level
			self:StopDragResizing();
		end
	end);
end

function DeveloperConsoleMixin:StopDragResizing()
	if self.resizingTicker then
		self.resizingTicker:Cancel();
		self.resizingTicker = nil;
	end
end

function DeveloperConsoleMixin:ExecuteCommand(text)
	ConsoleExec(text, true);
	self:AddMessage(("> %s"):format(text:gsub("\n", " > ")), Enum.ConsoleColorType.InputColor);
	if self.commandCircularBuffer:GetEntryAtIndex(1) ~= text then
		self.commandCircularBuffer:PushFront(text);
		table.insert(self.savedVars.commandHistory, text);
	end
	self:ResetCommandHistoryIndex();
end

function DeveloperConsoleMixin:OnEditBoxTextChanged(text)
	self:ValidateHeight();

	if self.ignoreNextTextChange then
		self.ignoreNextTextChange = nil;
		return;
	end

	local commandText, startPosition = self:FindBestEditCommand();
	self.AutoComplete:SetText(commandText, self.EditBox:GetCursorPosition() - (startPosition - 1));
end

function DeveloperConsoleMixin:OnEditBoxCursorChanged(text)
	if self.ignoreNextTextChange then
		return;
	end
	self:OnEditBoxTextChanged(text);
end

function DeveloperConsoleMixin:OnEditBoxArrowPressed(key)
	if key == "UP" then
		if self:GetCommandHistoryIndex() + 1 <= self.commandCircularBuffer:GetNumElements() then
			self:SetCommandHistoryIndex(self:GetCommandHistoryIndex() + 1);
			self.EditBox:SetText(self.commandCircularBuffer:GetEntryAtIndex(self:GetCommandHistoryIndex()));
		end
	elseif key == "DOWN" then
		if self:GetCommandHistoryIndex() - 1 > 0 then
			self:SetCommandHistoryIndex(self:GetCommandHistoryIndex() - 1);
			self.EditBox:SetText(self.commandCircularBuffer:GetEntryAtIndex(self:GetCommandHistoryIndex()));
		end
	end
end

function DeveloperConsoleMixin:OnEditBoxPageUpPressed()
	self.MessageFrame:PageUp();
end

function DeveloperConsoleMixin:OnEditBoxPageDownPressed()
	self.MessageFrame:PageDown();
end

function DeveloperConsoleMixin:OnEditBoxTabPressed()
	if IsControlKeyDown() then
		C_Console.PrintAllMatchingCommands((self:FindBestEditCommand()));
	else
		if IsShiftKeyDown() then
			if self.AutoComplete:PreviousEntry() then
				return;
			end
		else
			if self.AutoComplete:GetSelectedText() == self:FindBestEditCommand() then
				if self.AutoComplete:NextEntry() then
					return;
				end
			end
		end

		self:SetAutoCompleteText(self.AutoComplete:GetSelectedText(), true);
	end
end

do
	local function Matches(pattern, text)
		if #pattern == 0 then
			return true;
		end
		return text:lower():match(pattern);
	end

	function DeveloperConsoleMixin:OnFilterEditBoxTextChanged(text)
		if self.filterText == text then
			return;
		end
		self.filterText = text;

		self.MessageFrame:Clear();
		text = text:lower();

		self.filteringCoroutine = coroutine.create(function()
			local numMessages = #self.savedVars.messageHistory;
			self.Filters.ProgressBar:SetMinMaxSmoothedValue(0, numMessages);
			self.Filters.ProgressBar:ResetSmoothedValue(0);
			self.Filters.ProgressBar.FadeAnim:Play();
			for i = numMessages, 1, -1 do
				self:CheckFilterCoroutineYield();
				self.Filters.ProgressBar:SetSmoothedValue(numMessages - i + 1);

				local messageInfo = self.savedVars.messageHistory[i];
			
				local success, matched = pcall(Matches, text, messageInfo[1]);
				if success and matched then
					local message, colorType = unpack(messageInfo);
					local color = C_Console.GetColorFromType(colorType);
					local r, g, b = color:GetRGB();
					self.MessageFrame:BackFillMessage(message, r, g, b, colorType);
					self:UpdateScrollbar();
				end
			end
		end);

		local extendTime = true;
		self:StepFilteringCoroutine(extendTime);
	end
end

function DeveloperConsoleMixin:OnUpdate()
	self:StepFilteringCoroutine();
end

function DeveloperConsoleMixin:StepFilteringCoroutine(extendedTime)
	if self.filteringCoroutine then
		local TIME_PER_FRAME_SEC = 0.0015;
		local EXTENDED_TIME_PER_FRAME_SEC = TIME_PER_FRAME_SEC * 10;
		self.filteringCoroutineEndTime = GetTimePreciseSec() + (extendedTime and EXTENDED_TIME_PER_FRAME_SEC or TIME_PER_FRAME_SEC);

		coroutine.resume(self.filteringCoroutine);
		if coroutine.status(self.filteringCoroutine) == "dead" then
			self.filteringCoroutine = nil;
			self.Filters.ProgressBar.FadeAnim:Play(true);
		end
	end
end

function DeveloperConsoleMixin:CheckFilterCoroutineYield()
	if GetTimePreciseSec() > self.filteringCoroutineEndTime then
		return coroutine.yield();
	end
end

function DeveloperConsoleMixin:SetAutoCompleteText(newCommand, keepAutoComplete)
	local startPosition, endPosition, text = self:FindBestEditCommandPositions();
	local currentText = self.EditBox:GetText();
	local newTextStart = currentText:sub(1, startPosition - 1);
	local newTextEnd = currentText:sub(endPosition + 1);
	local commandArguments = text:match("^%S+%s+(.+)") or "";
	local newText = newTextStart .. newCommand .. " " .. commandArguments .. newTextEnd;

	self.ignoreNextTextChange = keepAutoComplete;
	self.EditBox:SetText(newText);
	self.EditBox:SetCursorPosition(startPosition + #newCommand + 1 + #commandArguments);
end

function DeveloperConsoleMixin:FindBestEditCommandPositions()
	local cursorPosition = self.EditBox:GetCursorPosition();

	local startPosition = 0;
	local currentText = self.EditBox:GetText();
	local text, rest = currentText:match("^(.-)\n(.*)$");
	while text do
		local endPosition = startPosition + #text;
		if cursorPosition >= startPosition and cursorPosition <= endPosition then
			return startPosition + 1, endPosition, text;
		end
		startPosition = endPosition + 1;
		if not rest then
			break;
		end
		text, rest = rest:match("^(.-)\n(.*)$");
	end

	return startPosition + 1, #currentText, currentText:sub(startPosition + 1);
end

function DeveloperConsoleMixin:FindBestEditCommand()
	local startPosition, endPosition, text = self:FindBestEditCommandPositions();
	return text:match("^(%S+)") or text, startPosition, endPosition, text;
end

function DeveloperConsoleMixin:ResetCommandHistoryIndex()
	self.commandHistoryIndex = nil;
end

function DeveloperConsoleMixin:GetCommandHistoryIndex()
	return self.commandHistoryIndex or 0;
end

function DeveloperConsoleMixin:SetCommandHistoryIndex(commandHistoryIndex)
	self.commandHistoryIndex = commandHistoryIndex;
end

function DeveloperConsoleMixin:HasSetCommandHistoryIndex()
	return self.commandHistoryIndex ~= nil;
end